@@ -66,11 +66,14 @@ static const char *
 version_and_features(void)
 {
     static char buf[256];
-    snprintf(buf, sizeof(buf), "version: %s %cpgo %cime %cgraphemes %cassertions",
+    snprintf(buf, sizeof(buf),
+             "version: %s %cpgo %cime %cgraphemes %cfractional-scaling %ccursor-shape %cassertions",
              FOOT_VERSION,
              feature_pgo() ? '+' : '-',
              feature_ime() ? '+' : '-',
              feature_graphemes() ? '+' : '-',
+             feature_fractional_scaling() ? '+' : ':',
+             feature_cursor_shape() ? '+' : '-',
              feature_assertions() ?  '+' : '-');
     return buf;
 }
@@ -371,16 +374,19 @@ main(int argc, char *const *argv)
         const char *xdg_runtime = getenv("XDG_RUNTIME_DIR");
         if (xdg_runtime != NULL) {
             const char *wayland_display = getenv("WAYLAND_DISPLAY");
-            if (wayland_display != NULL)
+            if (wayland_display != NULL) {
                 snprintf(addr.sun_path, sizeof(addr.sun_path),
                          "%s/foot-%s.sock", xdg_runtime, wayland_display);
-            else
+                connected = (connect(fd, (const struct sockaddr *)&addr, sizeof(addr)) == 0);
+            }
+            if (!connected) {
+                LOG_WARN("%s: failed to connect, will now try %s/foot.sock",
+                         addr.sun_path, xdg_runtime);
                 snprintf(addr.sun_path, sizeof(addr.sun_path),
                          "%s/foot.sock", xdg_runtime);
-
-            if (connect(fd, (const struct sockaddr *)&addr, sizeof(addr)) == 0)
-                connected = true;
-            else
+                connected = (connect(fd, (const struct sockaddr *)&addr, sizeof(addr)) == 0);
+            }
+            if (!connected)
                 LOG_WARN("%s: failed to connect, will now try /tmp/foot.sock", addr.sun_path);
         }
 
