@@ -37,3 +37,21 @@ static inline bool feature_graphemes(void)
     return false;
 #endif
 }
+
+static inline bool feature_fractional_scaling(void)
+{
+#if defined(HAVE_FRACTIONAL_SCALE)
+    return true;
+#else
+    return false;
+#endif
+}
+
+static inline bool feature_cursor_shape(void)
+{
+#if defined(HAVE_CURSOR_SHAPE)
+    return true;
+#else
+    return false;
+#endif
+}
