@@ -736,12 +736,14 @@ bool term_to_slave(struct terminal *term, const void *data, size_t len);
 bool term_paste_data_to_slave(
     struct terminal *term, const void *data, size_t len);
 
+bool term_fractional_scaling(const struct terminal *term);
 bool term_update_scale(struct terminal *term);
 bool term_font_size_increase(struct terminal *term);
 bool term_font_size_decrease(struct terminal *term);
 bool term_font_size_reset(struct terminal *term);
 bool term_font_dpi_changed(struct terminal *term, float old_scale);
 void term_font_subpixel_changed(struct terminal *term);
+int term_font_baseline(const struct terminal *term);
 
 int term_pt_or_px_as_pixels(
     const struct terminal *term, const struct pt_or_px *pt_or_px);
